@@ -32,7 +32,9 @@
 #include "dwg.h"
 #include "decode.h"
 #include "free.h"
+#include "classes.h"
 #include "hash.h"
+#include "free.h"
 
 static unsigned int loglevel;
 #ifdef USE_TRACING
@@ -43,7 +45,8 @@ static int env_var_checked_p;
 
 /* the current version per spec block */
 static unsigned int cur_ver = 0;
-static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
+static Bit_Chain pdat = { NULL, 0, 0, 0, 0, R_INVALID };
+static BITCODE_BL rcount1, rcount2;
 
 /*--------------------------------------------------------------------------------
  * MACROS
@@ -59,16 +62,28 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
     ptr = NULL;                                                               \
   }
 
+#undef UNTIL
 #undef SINCE
+#undef PRE
+#undef VERSIONS
+#undef VERSION
+#define UNTIL(v)                                                              \
+  if (dat->from_version <= v)
 #define SINCE(v)                                                              \
-  cur_ver = v;                                                                \
   if (dat->from_version >= v)
+#define PRE(v)                                                                \
+  if (dat->from_version < v)
+#define VERSIONS(v1, v2)                                                      \
+  if (dat->from_version >= v1 && dat->from_version <= v2)
+#define VERSION(v)                                                            \
+  if (dat->from_version == v)
 
 #define VALUE(value, type, dxf)
 #define VALUE_RC(value, dxf) VALUE (value, RC, dxf)
 #define VALUE_RS(value, dxf) VALUE (value, RS, dxf)
 #define VALUE_RL(value, dxf) VALUE (value, RL, dxf)
 #define VALUE_RD(value, dxf) VALUE (value, RD, dxf)
+#define VALUE_BD(value, dxf) VALUE (value, BD, dxf)
 
 #define FIELD(name, type)                                                     \
   {                                                                           \
@@ -81,6 +96,9 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
 #define FIELD_CAST(name, type, cast, dxf)                                     \
   {                                                                           \
   }
+#define SUB_FIELD_CAST(o, name, type, cast, dxf)                              \
+  {                                                                           \
+  }
 #define FIELD_VALUE(name) _obj->name
 #define SUB_FIELD(o, nam, type, dxf) FIELD (_obj->o.nam, type)
 
@@ -90,13 +108,10 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
   VALUE_HANDLE (_obj->o.nam, nam, code, dxf)
 // compare to dwg_decode_handleref_with_code: not all refs are stored in the
 // object_ref vector, like relative ptrs and NULL.
-// But indxf skip the NULL HDL, it is global and shared there.
+// But with INDXF skip the NULL HDL, it is global and shared there.
 // obj is the relative base object here and there.
 #define VALUE_HANDLE(ref, nam, _code, dxf)                                    \
-  if (ref                                                                     \
-      && !(dat->opts & DWG_OPTS_INDXF && ref->handleref.size == 0             \
-           && ref->absolute_ref == 0 && !ref->obj)                            \
-      && !(ref->handleref.size || (obj && ref->handleref.code > 5)))          \
+  if (ref && !ref->handleref.is_global)                                       \
     {                                                                         \
       free (ref);                                                             \
       ref = NULL;                                                             \
@@ -129,6 +144,9 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
   }
 #define FIELD_T(name, dxf) FIELD_TV (name, dxf)
 #define FIELD_BINARY(name, len, dxf) FIELD_TV (name, dxf)
+#define FIELD_T16(name, dxf) FIELD_TV (name, dxf)
+#define FIELD_TU16(name, dxf) FIELD_TV (name, dxf)
+#define FIELD_T32(name, dxf) FIELD_TV (name, dxf)
 #define FIELD_BT(name, dxf) FIELD (name, BT);
 #define FIELD_4BITS(name, dxf)                                                \
   {                                                                           \
@@ -139,7 +157,7 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
 #define FIELD_DD(name, _default, dxf)                                         \
   {                                                                           \
   }
-#define FIELD_2DD(name, d1, d2, dxf)                                          \
+#define FIELD_2DD(name, def, dxf)                                             \
   {                                                                           \
   }
 #define FIELD_3DD(name, def, dxf)                                             \
@@ -166,16 +184,21 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
 #define FIELD_3DPOINT(name, dxf)                                              \
   {                                                                           \
   }
+#define FIELD_2PT_TRACE(name, type, dxf)                                      \
+  {                                                                           \
+  }
+#define FIELD_3PT_TRACE(name, type, dxf)                                      \
+  {                                                                           \
+  }
 #define FIELD_TIMEBLL(name, dxf)
 #define FIELD_TIMERLL(name, dxf)
-// indxf does not assign names yet
-#define FIELD_CMC(color, dxf1, dxf2)                                          \
-  if (!(dat->opts & DWG_OPTS_INDXF)) {                                        \
+#define FIELD_CMC(color, dxf)                                                 \
+  {                                                                           \
     FIELD_T (color.name, 0);                                                  \
     FIELD_T (color.book_name, 0);                                             \
   }
-#define SUB_FIELD_CMC(o, color, dxf1, dxf2)                                   \
-  if (!(dat->opts & DWG_OPTS_INDXF)) {                                        \
+#define SUB_FIELD_CMC(o, color, dxf)                                          \
+  {                                                                           \
     VALUE_TV (_obj->o.color.name, 0);                                         \
     VALUE_TV (_obj->o.color.book_name, 0);                                    \
   }
@@ -190,7 +213,7 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
         FIELD_##type (nam[vcount], dxf);                                      \
     }                                                                         \
   FIELD_TV (nam, dxf);
-#define FIELD_VECTOR_T(name, size, dxf)                                       \
+#define FIELD_VECTOR_T(name, type, size, dxf)                                 \
   FIELD_VECTOR_N (name, TV, _obj->size, dxf)
 #define FIELD_VECTOR(name, type, size, dxf)                                   \
   FIELD_VECTOR_N (name, type, _obj->size, dxf)
@@ -211,6 +234,22 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
     }
 #define HANDLE_VECTOR(name, sizefield, code, dxf)                             \
   HANDLE_VECTOR_N (name, FIELD_VALUE (sizefield), code, dxf)
+#define SUB_HANDLE_VECTOR(o,name, sizefield, code, dxf)                       \
+  if (_obj->o.name && _obj->o.sizefield)                                      \
+    {                                                                         \
+      for (vcount = 0; vcount < (BITCODE_BL)_obj->o.sizefield; vcount++)      \
+        {                                                                     \
+          SUB_FIELD_HANDLE (o,name[vcount], code, dxf);                       \
+        }                                                                     \
+      FREE_IF (_obj->o.name)                                                  \
+    }
+#define SUB_FIELD_VECTOR(o, nam, sizefield, type, dxf)                        \
+  if (_obj->o.sizefield && _obj->o.nam)                                       \
+    {                                                                         \
+      for (vcount = 0; vcount < (BITCODE_BL) (_obj->o.sizefield); vcount++)   \
+        SUB_FIELD_##type (o, nam[vcount], dxf);                               \
+    }                                                                         \
+  FREE_IF (_obj->o.nam)
 
 #define FIELD_NUM_INSERTS(num_inserts, type, dxf)
 #define FIELD_XDATA(name, size) dwg_free_xdata (_obj, _obj->size)
@@ -233,7 +272,7 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
 #define XDICOBJHANDLE(code)                                                   \
   SINCE (R_2004)                                                              \
   {                                                                           \
-    if (!obj->tio.object->xdic_missing_flag)                                  \
+    if (!obj->tio.object->is_xdic_missing)                                  \
       {                                                                       \
         VALUE_HANDLE (obj->tio.object->xdicobjhandle, xdicobjhandle, code,    \
                       0);                                                     \
@@ -246,7 +285,7 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
 #define ENT_XDICOBJHANDLE(code)                                               \
   SINCE (R_2004)                                                              \
   {                                                                           \
-    if (!ent->xdic_missing_flag)                                              \
+    if (!ent->is_xdic_missing)                                              \
       {                                                                       \
         VALUE_HANDLE (ent->xdicobjhandle, xdicobjhandle, code, 0);            \
       }                                                                       \
@@ -261,14 +300,11 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
 #define END_STRING_STREAM
 #define START_HANDLE_STREAM
 
-static int dwg_free_UNKNOWN_ENT (Bit_Chain *restrict dat,
-                                 Dwg_Object *restrict obj);
-static int dwg_free_UNKNOWN_OBJ (Bit_Chain *restrict dat,
-                                 Dwg_Object *restrict obj);
 
 #define DWG_ENTITY(token)                                                     \
-  static int dwg_free_##token##_private (Bit_Chain *restrict dat,             \
-                                         Dwg_Object *restrict obj);           \
+  static int dwg_free_##token##_private                                       \
+    (Bit_Chain *dat, Bit_Chain *hdl_dat,                                      \
+     Bit_Chain *str_dat, Dwg_Object *restrict obj);                           \
                                                                               \
   static int dwg_free_##token (Bit_Chain *restrict dat,                       \
                                Dwg_Object *restrict obj)                      \
@@ -278,7 +314,7 @@ static int dwg_free_UNKNOWN_OBJ (Bit_Chain *restrict dat,
       {                                                                       \
         LOG_HANDLE ("Free entity " #token " [%d]\n", obj->index)              \
         if (obj->tio.entity->tio.token)                                       \
-          error = dwg_free_##token##_private (dat, obj);                      \
+          error = dwg_free_##token##_private (dat, dat, dat, obj);            \
                                                                               \
         dwg_free_common_entity_data (obj);                                    \
         dwg_free_eed (obj);                                                   \
@@ -291,26 +327,28 @@ static int dwg_free_UNKNOWN_OBJ (Bit_Chain *restrict dat,
     obj->parent = NULL;                                                       \
     return error;                                                             \
   }                                                                           \
-  static int dwg_free_##token##_private (Bit_Chain *restrict dat,             \
-                                         Dwg_Object *restrict obj)            \
+  static int dwg_free_##token##_private                                       \
+    (Bit_Chain *dat, Bit_Chain *hdl_dat,                                      \
+     Bit_Chain *str_dat, Dwg_Object *restrict obj)                            \
   {                                                                           \
-    BITCODE_BL vcount, rcount1, rcount2, rcount3, rcount4;                    \
+    BITCODE_BL vcount, rcount3, rcount4;                                      \
     Dwg_Entity_##token *ent, *_obj;                                           \
     Dwg_Object_Entity *_ent;                                                  \
-    Bit_Chain *hdl_dat = dat;                                                 \
-    Bit_Chain *str_dat = dat;                                                 \
     Dwg_Data *dwg = obj->parent;                                              \
     int error = 0;                                                            \
     _ent = obj->tio.entity;                                                   \
+    if (!_ent)                                                                \
+      return 0;                                                               \
     _obj = ent = _ent->tio.token;
 
 #define DWG_ENTITY_END                                                        \
-  return error;                                                               \
+    return error;                                                             \
   }
 
 #define DWG_OBJECT(token)                                                     \
-  static int dwg_free_##token##_private (Bit_Chain *restrict dat,             \
-                                         Dwg_Object *restrict obj);           \
+  static int dwg_free_##token##_private                                       \
+    (Bit_Chain *dat, Bit_Chain *hdl_dat,                                      \
+     Bit_Chain *str_dat, Dwg_Object *restrict obj);                           \
                                                                               \
   static int dwg_free_##token (Bit_Chain *restrict dat,                       \
                                Dwg_Object *restrict obj)                      \
@@ -321,7 +359,8 @@ static int dwg_free_UNKNOWN_OBJ (Bit_Chain *restrict dat,
       {                                                                       \
         _obj = obj->tio.object->tio.token;                                    \
         LOG_HANDLE ("Free object " #token " [%d]\n", obj->index)              \
-        error = dwg_free_##token##_private (dat, obj);                        \
+        error = dwg_free_##token##_private (dat, dat, dat, obj);              \
+        dwg_free_common_object_data (obj);                                    \
         dwg_free_eed (obj);                                                   \
         FREE_IF (_obj);                                                       \
         FREE_IF (obj->tio.object);                                            \
@@ -330,13 +369,12 @@ static int dwg_free_UNKNOWN_OBJ (Bit_Chain *restrict dat,
     return error;                                                             \
   }                                                                           \
                                                                               \
-  static int dwg_free_##token##_private (Bit_Chain *restrict dat,             \
-                                         Dwg_Object *restrict obj)            \
+  static int dwg_free_##token##_private                                       \
+    (Bit_Chain *dat, Bit_Chain *hdl_dat,                                      \
+     Bit_Chain *str_dat, Dwg_Object *restrict obj)                            \
   {                                                                           \
-    BITCODE_BL vcount, rcount1, rcount2, rcount3, rcount4;                    \
+    BITCODE_BL vcount, rcount3, rcount4;                                      \
     Dwg_Object_##token *_obj;                                                 \
-    Bit_Chain *hdl_dat = dat;                                                 \
-    Bit_Chain *str_dat = dat;                                                 \
     Dwg_Data *dwg = obj->parent;                                              \
     int error = 0;                                                            \
     if (!obj->tio.object)                                                     \
@@ -346,7 +384,7 @@ static int dwg_free_UNKNOWN_OBJ (Bit_Chain *restrict dat,
 /* obj itself is allocated via dwg->object[], dxfname is klass->dxfname or
  * static */
 #define DWG_OBJECT_END                                                        \
-  return error;                                                               \
+    return error;                                                             \
   }
 
 static void
@@ -370,12 +408,30 @@ dwg_free_common_entity_data (Dwg_Object *obj)
 
   // clang-format off
   #include "common_entity_data.spec"
+  if (dat->from_version >= R_2007 && ent->color.flag & 0x40)
+    FIELD_HANDLE (color.handle, 0, 430);
   SINCE (R_13) {
   #include "common_entity_handle_data.spec"
   }
   // clang-format on
 }
 
+static void
+dwg_free_common_object_data (Dwg_Object *obj)
+{
+
+  Dwg_Data *dwg = obj->parent;
+  Bit_Chain *dat = &pdat;
+  Bit_Chain *hdl_dat = &pdat;
+  Dwg_Object_Object *_obj = obj->tio.object;
+  BITCODE_BL vcount;
+  int error = 0;
+
+  // clang-format off
+  #include "common_object_handle_data.spec"
+  // clang-format on
+}
+
 static void
 dwg_free_xdata (Dwg_Object_XRECORD *obj, int size)
 {
@@ -415,27 +471,73 @@ dwg_free_eed (Dwg_Object *obj)
 
 #include "dwg.spec"
 
-/* returns 1 if object could be freed and 0 otherwise
+// could be a hash or switch, but there are not that many DEBUGGING classes.
+// but a switch is fine, as we get all missing types in objects.inc, generated by regen-dynapi
+int
+dwg_free_variable_no_class (Dwg_Data *restrict dwg, Dwg_Object *restrict obj)
+{
+  Bit_Chain *dat = &pdat;
+
+#undef DWG_ENTITY
+#undef DWG_OBJECT
+#define FREE_NOCLASS(name) case DWG_TYPE_##name: return dwg_free_##name (dat, obj);
+#define DWG_ENTITY(name) FREE_NOCLASS (name)
+#define DWG_OBJECT(name) FREE_NOCLASS (name)
+
+  switch (obj->fixedtype)
+    {
+    #include "objects.inc"
+
+    case DWG_TYPE_FREED: break; // already freed
+    case DWG_TYPE_UNUSED:
+    case DWG_TYPE_ACDSRECORD:
+    case DWG_TYPE_ACDSSCHEMA:
+    case DWG_TYPE_NPOCOLLECTION:
+    case DWG_TYPE_RAPIDRTRENDERENVIRONMENT:
+    case DWG_TYPE_XREFPANELOBJECT:
+    default: LOG_ERROR ("Unhandled class %s, fixedtype %d in objects.inc",
+                        dwg_type_name (obj->fixedtype), (int)obj->fixedtype);
+    }
+
+#undef DWG_ENTITY
+#undef DWG_OBJECT
+#undef FREE_NOCLASS
+
+  return DWG_ERR_UNHANDLEDCLASS;
+}
+
+/* returns error.
  */
-static int
-dwg_free_variable_type (Dwg_Data *restrict dwg, Dwg_Object *restrict obj)
+int dwg_free_variable_type (Dwg_Data *restrict dwg, Dwg_Object *restrict obj)
 {
   const int i = obj->type - 500;
   Dwg_Class *klass;
   Bit_Chain *dat = &pdat;
 
+  // no matching class
   if (i < 0 || i >= (int)dwg->num_classes)
-    return DWG_ERR_INVALIDTYPE;
+    {
+      LOG_WARN ("No class for %s type %d found", obj->name, obj->type);
+      return dwg_free_variable_no_class (dwg, obj);
+    }
 
   klass = &dwg->dwg_class[i];
   if (!klass || !klass->dxfname)
-    return DWG_ERR_INTERNALERROR;
+    {
+      LOG_WARN ("No class for %s type %d found", obj->name, obj->type);
+      return dwg_free_variable_no_class (dwg, obj);
+    }
 
   if (strNE (obj->dxfname, klass->dxfname))
     {
       LOG_ERROR ("Wrong %s.type %d for obj [%d]: != %s",  obj->dxfname, obj->type,
                  obj->index, klass->dxfname);
-      return DWG_ERR_INVALIDTYPE;
+      // But we know how to handle the UNKNOWN_* types
+      if (obj->fixedtype == DWG_TYPE_UNKNOWN_OBJ
+          || obj->fixedtype == DWG_TYPE_UNKNOWN_ENT)
+        return DWG_ERR_UNHANDLEDCLASS;
+      else
+        return dwg_free_variable_no_class (dwg, obj);
     }
 
   // global class dispatcher:
@@ -449,11 +551,46 @@ dwg_free_variable_type (Dwg_Data *restrict dwg, Dwg_Object *restrict obj)
   #undef WARN_UNSTABLE_CLASS
   // clang-format on
 
+  LOG_WARN ("No class for %s type %d found", obj->name, obj->type);
+  return dwg_free_variable_no_class (dwg, obj);
+}
+
+/* returns error */
+int dwg_free_variable_type_private (Dwg_Object *restrict obj)
+{
+  Dwg_Data *restrict dwg = obj->parent;
+  Bit_Chain *dat = &pdat;
+
+#undef DWG_ENTITY
+#undef DWG_OBJECT
+#define FREE_NOCLASS(name) case DWG_TYPE_##name: return dwg_free_##name##_private (dat, dat, dat, obj);
+#define DWG_ENTITY(name) FREE_NOCLASS (name)
+#define DWG_OBJECT(name) FREE_NOCLASS (name)
+
+  switch (obj->fixedtype)
+    {
+    #include "objects.inc"
+
+    case DWG_TYPE_FREED: break; // already freed
+    case DWG_TYPE_UNUSED:
+    case DWG_TYPE_ACDSRECORD:
+    case DWG_TYPE_ACDSSCHEMA:
+    case DWG_TYPE_NPOCOLLECTION:
+    case DWG_TYPE_RAPIDRTRENDERENVIRONMENT:
+    case DWG_TYPE_XREFPANELOBJECT:
+    default: LOG_ERROR ("Unhandled class %s, fixedtype %d in objects.inc",
+                        dwg_type_name (obj->fixedtype), (int)obj->fixedtype);
+    }
+
+#undef DWG_ENTITY
+#undef DWG_OBJECT
+#undef FREE_NOCLASS
+
   return DWG_ERR_UNHANDLEDCLASS;
 }
 
 // using the global dat
-void
+EXPORT void
 dwg_free_object (Dwg_Object *obj)
 {
   int error = 0;
@@ -663,11 +800,11 @@ dwg_free_object (Dwg_Object *obj)
     case DWG_TYPE_DIMSTYLE:
       dwg_free_DIMSTYLE (dat, obj);
       break;
-    case DWG_TYPE_VPORT_ENTITY_CONTROL:
-      dwg_free_VPORT_ENTITY_CONTROL (dat, obj);
+    case DWG_TYPE_VX_CONTROL:
+      dwg_free_VX_CONTROL (dat, obj);
       break;
-    case DWG_TYPE_VPORT_ENTITY_HEADER:
-      dwg_free_VPORT_ENTITY_HEADER (dat, obj);
+    case DWG_TYPE_VX_TABLE_RECORD:
+      dwg_free_VX_TABLE_RECORD (dat, obj);
       break;
     case DWG_TYPE_GROUP:
       dwg_free_GROUP (dat, obj);
@@ -731,37 +868,371 @@ dwg_free_object (Dwg_Object *obj)
               dwg_free_UNKNOWN_OBJ (dat, obj);
         }
     }
-  /* With this importer the dxfname is dynamic, just the name is const */
-  if (dwg->opts & DWG_OPTS_INDXF)
+  /* With indxf the dxfname is dynamic, just the name is const */
+  if (dwg->opts & DWG_OPTS_IN)
     FREE_IF (obj->dxfname);
+  /* With injson even the name is dynamic */
+  if (dwg->opts & DWG_OPTS_INJSON)
+    FREE_IF (obj->name);
   obj->type = DWG_TYPE_FREED;
 }
 
+/* Needed when we cast types.
+   By fixedtype, not dxfname.
+ */
+EXPORT void
+dwg_free_object_private (Dwg_Object *obj)
+{
+  int error = 0;
+  long unsigned int j;
+  Dwg_Data *dwg;
+  Bit_Chain *dat = &pdat;
+
+  if (obj && obj->parent)
+    {
+      dwg = obj->parent;
+      dat->version = dwg->header.version;
+      dat->from_version = dwg->header.from_version;
+    }
+  else
+    return;
+  if (obj->type == DWG_TYPE_FREED || obj->tio.object == NULL)
+    return;
+
+  switch (obj->type)
+    {
+    case DWG_TYPE_TEXT:
+      dwg_free_TEXT_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_ATTRIB:
+      dwg_free_ATTRIB_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_ATTDEF:
+      dwg_free_ATTDEF_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_BLOCK:
+      dwg_free_BLOCK_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_ENDBLK:
+      dwg_free_ENDBLK_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_SEQEND:
+      dwg_free_SEQEND_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_INSERT:
+      dwg_free_INSERT_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_MINSERT:
+      dwg_free_MINSERT_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_VERTEX_2D:
+      dwg_free_VERTEX_2D_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_VERTEX_3D:
+      dwg_free_VERTEX_3D_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_VERTEX_MESH:
+      dwg_free_VERTEX_MESH_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_VERTEX_PFACE:
+      dwg_free_VERTEX_PFACE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_VERTEX_PFACE_FACE:
+      dwg_free_VERTEX_PFACE_FACE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_POLYLINE_2D:
+      dwg_free_POLYLINE_2D_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_POLYLINE_3D:
+      dwg_free_POLYLINE_3D_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_ARC:
+      dwg_free_ARC_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_CIRCLE:
+      dwg_free_CIRCLE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_LINE:
+      dwg_free_LINE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_DIMENSION_ORDINATE:
+      dwg_free_DIMENSION_ORDINATE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_DIMENSION_LINEAR:
+      dwg_free_DIMENSION_LINEAR_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_DIMENSION_ALIGNED:
+      dwg_free_DIMENSION_ALIGNED_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_DIMENSION_ANG3PT:
+      dwg_free_DIMENSION_ANG3PT_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_DIMENSION_ANG2LN:
+      dwg_free_DIMENSION_ANG2LN_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_DIMENSION_RADIUS:
+      dwg_free_DIMENSION_RADIUS_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_DIMENSION_DIAMETER:
+      dwg_free_DIMENSION_DIAMETER_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_POINT:
+      dwg_free_POINT_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE__3DFACE:
+      dwg_free__3DFACE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_POLYLINE_PFACE:
+      dwg_free_POLYLINE_PFACE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_POLYLINE_MESH:
+      dwg_free_POLYLINE_MESH_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_SOLID:
+      dwg_free_SOLID_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_TRACE:
+      dwg_free_TRACE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_SHAPE:
+      dwg_free_SHAPE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_VIEWPORT:
+      dwg_free_VIEWPORT_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_ELLIPSE:
+      dwg_free_ELLIPSE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_SPLINE:
+      dwg_free_SPLINE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_REGION:
+      dwg_free_REGION_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE__3DSOLID:
+      dwg_free__3DSOLID_private (dat, dat, dat, obj);
+      break; /* Check the type of the object */
+    case DWG_TYPE_BODY:
+      dwg_free_BODY_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_RAY:
+      dwg_free_RAY_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_XLINE:
+      dwg_free_XLINE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_DICTIONARY:
+      dwg_free_DICTIONARY_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_MTEXT:
+      dwg_free_MTEXT_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_LEADER:
+      dwg_free_LEADER_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_TOLERANCE:
+      dwg_free_TOLERANCE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_MLINE:
+      dwg_free_MLINE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_BLOCK_CONTROL:
+      dwg_free_BLOCK_CONTROL_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_BLOCK_HEADER:
+      dwg_free_BLOCK_HEADER_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_LAYER_CONTROL:
+      dwg_free_LAYER_CONTROL_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_LAYER:
+      dwg_free_LAYER_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_STYLE_CONTROL:
+      dwg_free_STYLE_CONTROL_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_STYLE:
+      dwg_free_STYLE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_LTYPE_CONTROL:
+      dwg_free_LTYPE_CONTROL_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_LTYPE:
+      dwg_free_LTYPE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_VIEW_CONTROL:
+      dwg_free_VIEW_CONTROL_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_VIEW:
+      dwg_free_VIEW_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_UCS_CONTROL:
+      dwg_free_UCS_CONTROL_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_UCS:
+      dwg_free_UCS_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_VPORT_CONTROL:
+      dwg_free_VPORT_CONTROL_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_VPORT:
+      dwg_free_VPORT_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_APPID_CONTROL:
+      dwg_free_APPID_CONTROL_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_APPID:
+      dwg_free_APPID_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_DIMSTYLE_CONTROL:
+      dwg_free_DIMSTYLE_CONTROL_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_DIMSTYLE:
+      dwg_free_DIMSTYLE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_VX_CONTROL:
+      dwg_free_VX_CONTROL_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_VX_TABLE_RECORD:
+      dwg_free_VX_TABLE_RECORD_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_GROUP:
+      dwg_free_GROUP_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_MLINESTYLE:
+      dwg_free_MLINESTYLE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_OLE2FRAME:
+      dwg_free_OLE2FRAME_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_DUMMY:
+      dwg_free_DUMMY_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_LONG_TRANSACTION:
+      dwg_free_LONG_TRANSACTION_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_LWPOLYLINE:
+      dwg_free_LWPOLYLINE_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_HATCH:
+      dwg_free_HATCH_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_XRECORD:
+      dwg_free_XRECORD_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_PLACEHOLDER:
+      dwg_free_PLACEHOLDER_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_OLEFRAME:
+      dwg_free_OLEFRAME_private (dat, dat, dat, obj);
+      break;
+#ifdef DEBUG_VBA_PROJECT
+    case DWG_TYPE_VBA_PROJECT:
+      dwg_free_VBA_PROJECT_private (dat, dat, dat, obj);
+      break;
+#endif
+    case DWG_TYPE_LAYOUT:
+      dwg_free_LAYOUT_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_PROXY_ENTITY:
+      dwg_free_PROXY_ENTITY_private (dat, dat, dat, obj);
+      break;
+    case DWG_TYPE_PROXY_OBJECT:
+      dwg_free_PROXY_OBJECT_private (dat, dat, dat, obj);
+      break;
+    default:
+      if (obj->type == obj->parent->layout_type
+          && obj->fixedtype == DWG_TYPE_LAYOUT)
+        {
+          SINCE (R_13)
+          {
+            dwg_free_LAYOUT_private (dat, dat, dat, obj); // XXX avoid double-free, esp. in eed
+          }
+        }
+      else
+        dwg_free_variable_type_private (obj);
+    }
+}
+
 static int
 dwg_free_header_vars (Dwg_Data *dwg)
 {
   Dwg_Header_Variables *_obj = &dwg->header_vars;
   Dwg_Object *obj = NULL;
   Bit_Chain *dat = &pdat;
+
   // clang-format off
   #include "header_variables.spec"
   // clang-format on
+
   return 0;
 }
 
 static int
 dwg_free_summaryinfo (Dwg_Data *dwg)
 {
-  struct Dwg_SummaryInfo *_obj = &dwg->summaryinfo;
+  Dwg_SummaryInfo *_obj = &dwg->summaryinfo;
   Dwg_Object *obj = NULL;
   Bit_Chain *dat = &pdat;
-  BITCODE_RL rcount1, rcount2;
+
   // clang-format off
   #include "summaryinfo.spec"
   // clang-format on
   return 0;
 }
 
+static int
+dwg_free_appinfo (Dwg_Data *dwg)
+{
+  Dwg_AppInfo *_obj = &dwg->appinfo;
+  Dwg_Object *obj = NULL;
+  Bit_Chain *dat = &pdat;
+
+  // clang-format off
+  #include "appinfo.spec"
+  // clang-format on
+  return 0;
+}
+static int
+dwg_free_filedeplist (Dwg_Data *dwg)
+{
+  Dwg_FileDepList *_obj = &dwg->filedeplist;
+  Dwg_Object *obj = NULL;
+  Bit_Chain *dat = &pdat;
+  BITCODE_RL vcount;
+
+  // clang-format off
+  #include "filedeplist.spec"
+  // clang-format on
+  return 0;
+}
+static int
+dwg_free_security (Dwg_Data *dwg)
+{
+  Dwg_Security *_obj = &dwg->security;
+  Dwg_Object *obj = NULL;
+  Bit_Chain *dat = &pdat;
+
+  // clang-format off
+  #include "security.spec"
+  // clang-format on
+  return 0;
+}
+
+static int
+dwg_free_acds (Dwg_Data *dwg)
+{
+  Dwg_AcDs *_obj = &dwg->acds;
+  Dwg_Object *obj = NULL;
+  Bit_Chain *dat = &pdat;
+  BITCODE_RL rcount3 = 0, rcount4, vcount;
+
+  // clang-format off
+  #include "acds.spec"
+  // clang-format on
+  return 0;
+}
+
 void
 dwg_free (Dwg_Data *dwg)
 {
@@ -792,7 +1263,6 @@ dwg_free (Dwg_Data *dwg)
           if (!dwg_obj_is_control (&dwg->object[i]))
             dwg_free_object (&dwg->object[i]);
         }
-      FREE_IF (dwg->header.section);
       dwg_free_header_vars (dwg);
       dwg_free_summaryinfo (dwg);
       FREE_IF (dwg->thumbnail.chain);
@@ -802,6 +1272,17 @@ dwg_free (Dwg_Data *dwg)
             FREE_IF (dwg->header.section_info[i].sections);
           FREE_IF (dwg->header.section_info);
         }
+      dwg_free_appinfo (dwg);
+      dwg_free_filedeplist (dwg);
+      dwg_free_security (dwg);
+      dwg_free_acds (dwg);
+
+      FREE_IF (dwg->vbaproject.unknown_bits);
+      FREE_IF (dwg->revhistory.histories);
+      FREE_IF (dwg->appinfohistory.unknown_bits);
+      //FREE_IF (dwg->objfreespace...);
+      FREE_IF (dwg->Template.description);
+      FREE_IF (dwg->header.section);
       for (i = 0; i < dwg->second_header.num_handlers; i++)
         FREE_IF (dwg->second_header.handlers[i].data);
       // auxheader has no strings
@@ -820,8 +1301,8 @@ dwg_free (Dwg_Data *dwg)
               if (dwg->header.from_version >= R_2007)
                 FREE_IF (dwg->dwg_class[i].dxfname_u);
             }
-          FREE_IF (dwg->dwg_class);
         }
+      FREE_IF (dwg->dwg_class);
       if (dwg->object_ref)
         {
           LOG_HANDLE ("free %d global refs\n", dwg->num_object_refs)
@@ -832,9 +1313,15 @@ dwg_free (Dwg_Data *dwg)
             }
         }
       FREE_IF (dwg->object_ref);
+      for (i = 0; i < dwg->num_acis_sab_hdl; ++i)
+        {
+          FREE_IF (dwg->acis_sab_hdl[i]);
+        }
+      FREE_IF (dwg->acis_sab_hdl);
       FREE_IF (dwg->object);
       if (dwg->object_map)
         hash_free (dwg->object_map);
+      dwg->num_objects = dwg->num_classes = dwg->num_object_refs = 0;
 #undef FREE_IF
     }
 }
